#!/usr/bin/perl -w

# Copyright 2016, 2017, 2020 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  See the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.

use 5.006;
use strict;
use warnings;
use Carp 'croak';
use List::Util 'min';
use Math::NumSeq::PlanePathDelta;
use Math::PlanePath::AlternatePaper;
use Math::PlanePath::KochCurve;
use Test::More;

use lib 'tools';
use lib '.';
use MyAlternate ':all';
use MyGraphs;

plan tests => 266;


#------------------------------------------------------------------------------
# Koch and PmOneBits

{
  my $path = Math::PlanePath::KochCurve->new;
  my $seq = Math::NumSeq::PlanePathDelta->new (planepath => 'KochCurve',
                                               delta_type => 'TDir6');
  my $bad = 0;
  foreach my $n (0 .. 4**6) {
    my $by_func = PmOneBits($n) % 6;
    my $by_path = $seq->ith($n);
    unless ($by_func == $by_path) {
      $bad++;
    }
  }
  is($bad, 0, 'Math::PlanePath::KochCurve vs PmOneBits()');
}


#------------------------------------------------------------------------------
# twin alternate area tree domnum

{
  require Algorithm::ChooseSubsets;
  require Graph::Maker::TwinAlternateAreaTree;
  my $k = 3;
  my $graph = Graph::Maker->new ('twin_alternate_area_tree',
                                 level => $k,
                                 undirected => 1);
  my $whole_domnum = MyGraphs::Graph_tree_domnum($graph);
  is ($whole_domnum, 3);

  my @vertices = $graph->vertices;
  is (scalar(@vertices), 8);

  my @conns = (0, TA_a($k), TA_c($k), TA_e($k));
  my @dom_initial = (0) x scalar(@vertices);
  is (scalar(@dom_initial), 8);

  foreach my $d (@conns) {
    $dom_initial[$d] = 1;
  }

  my $it = Algorithm::ChooseSubsets->new(\@vertices);
  my $min = @vertices;
  SET: while (my $set = $it->next) {
    my @dom = @dom_initial;
    foreach my $v (@$set) {
      foreach my $d ($v, $graph->neighbours($v)) {
        $dom[$d] = 1;
      }
    }

    foreach my $i (0 .. $#dom) {
      if (! $dom[$i]) { next SET; } # not dominating
    }
    my $set_size = @$set;
    $min = min($min, $set_size);
  }
  is ($min, $whole_domnum);
}

#------------------------------------------------------------------------------
# GRS

{
  my @GRS_samples = (1, 1, 1, -1, 1, 1, -1, 1, 1, 1, 1, -1, -1, -1, 1, -1, 1);
  foreach my $n (0 .. $#GRS_samples) {
    is (GRS($n), $GRS_samples[$n], "GRS $n");
  }
}

#------------------------------------------------------------------------------
# alternate graph no Hamiltonian path

foreach my $k (1 .. 5) {
  my $graph = alternate_Graph($k);

  # never a cycle
  ok (! MyGraphs::Graph_is_Hamiltonian($graph),
      "k=$k no Hamiltonian cycle");

  # path up to k=3 which is where a hanging square
  is (! MyGraphs::Graph_is_Hamiltonian($graph, type => 'path'),
      $k >= 3,
      "k=$k no Hamiltonian path");

  MyGraphs::Graph_delete_hanging_cycles($graph);

  # never a cycle
  ok (! MyGraphs::Graph_is_Hamiltonian($graph),
      "k=$k no Hamiltonian cycle, no hanging");

  is (! MyGraphs::Graph_is_Hamiltonian($graph, type => 'path'),
      $k >= 4,
      "k=$k no Hamiltonian path, no hanging");
}


#------------------------------------------------------------------------------
# TAtospine

{
  my @TAtospineDir_samples = (1, 2, 1, 1, 0, 2, 1, 2, 1, 2, 3, 1, 0);
  foreach my $i (0 .. $#TAtospineDir_samples) {
    my $n = $i + 1;
    is (TAtospineDir($n), $TAtospineDir_samples[$i], "TAtospineDir");
  }
}

#------------------------------------------------------------------------------
# TAVertexToUpper etc

is (TAVertexToUpper(0), 0b11);
is (TAVertexToUpper(1), 0b110);
is (TAVertexToUpper(2), undef);
is (TAVertexToUpper(0b00111), 0b11000);
is (TAVertexToUpper(0b10111), undef);

is (TAVertexToLower(0b11000), 0b00111);
is (TAVertexToLower(0b01000), undef);



#------------------------------------------------------------------------------
# Xnum etc

{
  # Xnum values
  my @Xnum_samples = (0, 1, 4, 5, 16, 17, 20, 21, 64, 65, 68, 69);
  foreach my $x (0 .. $#Xnum_samples) {
    is (Xnum($x), $Xnum_samples[$x], "Xnum x=$x");
  }
}
{
  # Gnum values
  my @Gnum_samples = (0, 2, 8, 10, 32, 34, 40, 42, 128, 130, 136, 138);
  foreach my $g (0 .. $#Gnum_samples) {
    is (Gnum($g), $Gnum_samples[$g], "Gnum g=$g");
  }
}
{
  # MfirstN values
  my @MfirstN_samples = (0,1,2,3,6,9,10,11,22,33,34,35,38,41,42,43,86,129,130,131,134);
  foreach my $s (0 .. $#MfirstN_samples) {
    is (MfirstN($s), $MfirstN_samples[$s], "MfirstN g=$s");
  }
}
{
  # MlastN values
  my @MlastN_samples = (0,1,4,7,16,27,28,31,64,107,108,111,112,123,124,127,256,427,428,431);
  foreach my $s (0 .. $#MlastN_samples) {
    is (MlastN($s), $MlastN_samples[$s], "MlastN g=$s");
  }
}

#------------------------------------------------------------------------------
# triangle graph

# return a Graph.pm of a triangle with height $n
#                  *  n=4
#                  |
#              *---*
#              |   |
#          *---*---*
#          |   |   |
#      *---*---*---*
#      |   |   |   |
#  *---*---*---*---*  n=0
#
sub triangle_Graph {
  my ($n) = @_;
  require Graph::Undirected;
  my $graph = Graph::Undirected->new;
  $graph->set_graph_attribute('vertex_name_type','xy');
  foreach my $x (0 .. $n) {
    foreach my $y (0 .. $x) {
      $graph->add_vertex("$x,$y");
      if ($x > $y) { $graph->add_edge("$x,$y", ($x-1).",$y"); }
      if ($y > 0) { $graph->add_edge("$x,$y", "$x,".($y-1)); }
    }
  }
  return $graph;
}

# triangle_Graph(0)->MyGraphs::Graph_view;
# triangle_Graph(3)->MyGraphs::Graph_view;

{
  # Wtriangle values
  my @Wtriangle_samples = (0, 4, 28, 108, 308, 728, 1512);
  foreach my $n (0 .. $#Wtriangle_samples) {
    my $graph = triangle_Graph($n);
    my $W = MyGraphs::Graph_Wiener_index($graph);
    is ($W, $Wtriangle_samples[$n], "triangle Wiener samples n=$n");
  }
}
{
  # WtriangleTerm00 values
  my @WtriangleTerm00_samples = (0, 3, 12, 30, 60, 105, 168);
  foreach my $n (0 .. $#WtriangleTerm00_samples) {
    my $graph = triangle_Graph($n);
    my $W = MyGraphs::Graph_Wiener_part_at_vertex($graph,"0,0");
    is ($W, $WtriangleTerm00_samples[$n], "triangle Wiener samples n=$n");
  }
}

#------------------------------------------------------------------------------
# pyramid graph

# return a Graph.pm of a pyramid with height $n
#
#                  *                  n=4
#                  |
#              *---*---*
#              |   |   |
#          *---*---*---*---*
#          |   |   |   |   |
#      *---*---*---*---*---*---*
#      |   |   |   |   |   |   |
#  *---*---*---*---*---*---*---*---*  n=0
#
sub pyramid_Graph {
  my ($n) = @_;
  require Graph::Undirected;
  my $graph = Graph::Undirected->new;
  $graph->set_graph_attribute('vertex_name_type','xy');
  foreach my $y (0 .. $n) {
    foreach my $x (-$n+$y .. $n-$y) {
      $graph->add_vertex("$x,$y");
      if ($x > -$n+$y) { $graph->add_edge("$x,$y", ($x-1).",$y"); }
      if ($y > 0) { $graph->add_edge("$x,$y", "$x,".($y-1)); }
    }
  }
  return $graph;
}

# pyramid_Graph(0)->MyGraphs::Graph_view;
# pyramid_Graph(3)->MyGraphs::Graph_view;

{
  # Wpyramid values
  my @Wpyramid_samples = (0, 9, 80, 354, 1104, 2779, 6048);
  foreach my $n (0 .. $#Wpyramid_samples) {
    my $graph = pyramid_Graph($n);
    my $W = MyGraphs::Graph_Wiener_index($graph);
    is ($W, $Wpyramid_samples[$n], "Wpyramid_samples n=$n");
  }
}
{
  # WpyramidTermN0 values
  my @WpyramidTermN0_samples = (0, 5, 23, 62, 130, 235, 385);
  foreach my $n (0 .. $#WpyramidTermN0_samples) {
    my $graph = pyramid_Graph($n);
    my $W = MyGraphs::Graph_Wiener_part_at_vertex($graph,"$n,0");
    is ($W, $WpyramidTermN0_samples[$n], "WpyramidTermN0_samples n=$n");
  }
}

#------------------------------------------------------------------------------
# Graphs

{
  # Wiener index values
  my @Wiener_samples = (1, 4, 20, 65, 272, 1022, 4768, 20780);
  foreach my $k (0 .. $#Wiener_samples) {
    my $graph = alternate_Graph($k);
    my $W = MyGraphs::Graph_Wiener_index($graph);
    is ($W, $Wiener_samples[$k], "Wiener samples k=$k");
  }
}

{
  # Diameter sample values
  my @Diameter_samples = (1,2,4,5, 7,9, 15,17, 31,33);
  foreach my $k (0 .. $#Diameter_samples) {
    is (Diameter($k), $Diameter_samples[$k], "Diameter samples k=$k");
  }
}

# Diameter vs function
foreach my $k (0 .. 8) {
  my $graph = alternate_Graph($k);
  my $diameter = $graph->diameter;
  is (Diameter($k), $diameter, "Diameter by Graph k=$k");
}



#------------------------------------------------------------------------------
# MyAlternateConf

{
  my $s = MyAlternateConf->initial;
  is($s->num_segs, 1);
  ok($s->equals($s));

  foreach my $bit (0,1) {
    my $e = $s->expand($bit);
    is($e->num_segs, 2);
  }
  ok(! $s->expand(0)->equals($s));
  ok(! $s->expand(1)->equals($s));

  # ok(  $s->is_boundary(0, $R_square));
  # ok(! $s->is_boundary(1, $R_square));
  # ok(! $s->is_boundary(0, $L_square));
  # ok(  $s->is_boundary(1, $L_square));

  # foreach my $c ($R_square, $L_square) {
  #     $c->print;
  #   foreach my $bit (0, 1) {
  #     $c->expand($bit)->print;
  #     ok( $c->is_subset( $c->expand($bit) ),
  #       "$c->{name} bit=$bit");
  #   }
  # }

  my $B = $R_square_segs->union($L_square_segs)->set_name('B squares');
  # $B->print;
  foreach my $bit (0, 1) {
    # $B->expand($bit)->print;
    ok( $B->is_subset( $B->expand($bit) ),  "$B->{name} bit=$bit");
  }
}

# MyAlternateConf->initial->print;
# MyAlternateConf->initial->expand(0)->print;
# MyAlternateConf->initial->expand(1)->print;
# MyAlternateConf->initial->expand(0)->expand(0)->print;
# MyAlternateConf->initial->expand(0)->expand(1)->print;
# MyAlternateConf->initial->expand(1)->expand(0)->print;
# MyAlternateConf->initial->expand(1)->expand(1)->print;


#------------------------------------------------------------------------------
# dir()

{
  my @dir_samples = (1,0,-1,0,1,2,1,0,1,0,-1,-2,-1,0,-1,0,1,0,-1,0);
  foreach my $i (0 .. $#dir_samples) {
    my $n = $i+1;
    is (dir($n), $dir_samples[$i]);
  }
}


#------------------------------------------------------------------------------
# turn()

{
  my @turn_samples = (1,-1,-1,1,1,1,-1,-1,1,-1,-1,-1,1,1,-1,1,1,-1,-1,1);
  foreach my $i (0 .. $#turn_samples) {
    my $n = $i+1;
    is (turn($n), $turn_samples[$i]);
  }
}


#------------------------------------------------------------------------------
# various measures

{
  my @A_samples = (0, 0, 0, 1, 3, 9, 21, 49, 105, 225, 465, 961);
  foreach my $k (0 .. $#A_samples) {
    is (A($k), $A_samples[$k], "A");
  }
}
{
  my @AR_samples = (0, 0, 0, 0, 2, 3, 12, 21, 56, 105, 240, 465);
  foreach my $k (0 .. $#AR_samples) {
    is (AR($k), $AR_samples[$k], "AR");
  }
}
{
  my @AL_samples = (0, 0, 0, 1, 1, 6, 9, 28, 49, 120, 225, 496);
  foreach my $k (0 .. $#AL_samples) {
    is (AL($k), $AL_samples[$k], "AL");
  }
}

__END__

#------------------------------------------------------------------------------
# NalternateToPlus()

{
  my @NalternateToPlus_samples = (0,1,3,5,7,6,11,13,15,14,12,10,23,22,27,29);
  foreach my $n (0 .. $#NalternateToPlus_samples) {
    is (NalternateToPlus($n), $NalternateToPlus_samples[$n], "NalternateToPlus n=$n");
  }
}


#------------------------------------------------------------------------------
# Rpred()

{
  my $path = AlternatePaper_path();
  my $bad = 0;
  foreach my $n (0 .. 1024) {
    my $by_func = Rpred($n) ?1:0;
    my $by_path = $path->_UNDOCUMENTED__n_segment_is_right_boundary($n) ?1:0;
    unless ($by_func == $by_path) {
      $bad++;
    }
  }
  is($bad, 0);
}


#------------------------------------------------------------------------------
# various measures

{
  my @R_samples = (1, 2, 4, 8, 16, 28, 48, 84, 144, 244, 416, 708, 1200, 2036);
  foreach my $k (0 .. $#R_samples) {
    is (R($k), $R_samples[$k], "R");
  }
}
{
  my @L_samples = (1, 2, 4, 8, 12, 20, 36, 60, 100, 172, 292, 492, 836, 1420);
  foreach my $k (0 .. $#L_samples) {
    is (L($k), $L_samples[$k], "L");
  }
}
{
  my @B_samples = (2, 4, 8, 16, 28, 48, 84, 144, 244, 416, 708, 1200);
  foreach my $k (0 .. $#B_samples) {
    is (B($k), $B_samples[$k], "B");
  }
}

{
  my @RQ_samples = (1, 2, 3, 5, 9, 15, 25, 43, 73, 123, 209, 355, 601, 1019, 1729);
  foreach my $k (0 .. $#RQ_samples) {
    is (RQ($k), $RQ_samples[$k], "RQ");
  }
}
{
  foreach my $k (0 .. 20) {
    is (BQ3o($k), JA($k));
  }
}
{
  my @LQ_samples = (1, 1, 2, 4, 6, 10, 18, 30, 50, 86, 146, 246, 418, 710);
  foreach my $k (0 .. $#LQ_samples) {
    is (LQ($k), $LQ_samples[$k], "LQ");
  }
}

{
  #                 0  1  2  3  4
  my @JA_samples = (0, 0, 0, 1, 2, 3, 6, 11, 18, 31, 54, 91, 154, 263, 446, 755);
  foreach my $k (0 .. $#JA_samples) {
    is (JA($k), $JA_samples[$k], "JA");
  }
}
{
  #                 0  1  2  3  4
  my @JN_samples = (1, 2, 4, 7, 14, 28, 55, 110, 220, 439, 878);
  foreach my $k (0 .. $#JN_samples) {
    is (JN($k), $JN_samples[$k], "JN");
  }
}

{
  my @S_samples = (2, 3, 5, 9, 15, 25, 43, 73, 123, 209, 355, 601, 1019, 1729);
  foreach my $k (0 .. $#S_samples) {
    is (S($k), $S_samples[$k], "S");
  }
}
{
  my @P_samples = (2, 3, 5, 9, 16, 29, 54, 101, 190, 361, 690, 1325, 2558, 4961);
  foreach my $k (0 .. $#P_samples) {
    is (P($k), $P_samples[$k], "P");
  }
}
{
  my @Dpred_samples =
    ('', # 0
     '', # 1
     '', # 2
     '', # 3
     '7,11', # 4
     '7,11,14,15,19,21,22,25', # 5
     '7,11,14,15,19,21,22,25,28,30,31,35,38,39,42,43,44,45,49,50,53,57');
  foreach my $k (0 .. $#Dpred_samples) {
    my $got = join(',', grep {Dpred_k($k,$_)} 0 .. (1<<$k)-1);
    my $want = $Dpred_samples[$k];
    is($got, $want, "Dpred k=$k");
  }
}


#------------------------------------------------------------------------------
exit 0;
